"""
Type stubs for typed password fields.

These stubs tell type checkers that field constructors return primitive types,
enabling typed model definitions like:
    password: str = types.PasswordField()

At runtime, these are Field instances (descriptors), but type checkers see the primitives.

The return type is conditional on allow_null:
- allow_null=False (default) returns str
- allow_null=True returns str | None
"""

from collections.abc import Callable, Sequence
from typing import Any, Literal, overload

# PasswordField extends CharField with password-specific hashing
@overload
def PasswordField(
    *,
    required: bool = True,
    allow_null: Literal[True],
    default: Any = ...,
    choices: Any = None,
    db_column: str | None = None,
    db_collation: str | None = None,
    validators: Sequence[Callable[..., Any]] = (),
    error_messages: dict[str, str] | None = None,
    db_comment: str | None = None,
) -> str | None: ...
@overload
def PasswordField(
    *,
    required: bool = True,
    allow_null: Literal[False] = False,
    default: Any = ...,
    choices: Any = None,
    db_column: str | None = None,
    db_collation: str | None = None,
    validators: Sequence[Callable[..., Any]] = (),
    error_messages: dict[str, str] | None = None,
    db_comment: str | None = None,
) -> str: ...



"""
Typed field imports for better IDE and type checker support.

This module provides password-specific field classes with a companion .pyi
stub file that makes type checkers interpret field assignments as their
primitive Python types.

Usage:
    from plain.passwords.types import PasswordField

    @models.register_model
    class User(models.Model):
        email: str = types.EmailField()
        password: str = PasswordField()

This is optional - you can continue using untyped field definitions.
"""

from plain.passwords.models import PasswordField

__all__ = ["PasswordField"]



from __future__ import annotations

import gzip
from functools import cached_property
from pathlib import Path

from plain.exceptions import (
    ValidationError,
)
from plain.utils.deconstruct import deconstructible
from plain.utils.text import pluralize


@deconstructible
class MinimumLengthValidator:
    """
    Validate that the password is of a minimum length.
    """

    def __init__(self, min_length: int = 8) -> None:
        self.min_length = min_length

    def __call__(self, password: str) -> None:
        if len(password) < self.min_length:
            raise ValidationError(
                pluralize(
                    "This password is too short. It must contain at least "
                    "%(min_length)d character.",
                    "This password is too short. It must contain at least "
                    "%(min_length)d characters.",
                    self.min_length,
                ),
                code="password_too_short",
                params={"min_length": self.min_length},
            )


@deconstructible
class CommonPasswordValidator:
    """
    Validate that the password is not a common password.

    The password is rejected if it occurs in a provided list of passwords,
    which may be gzipped. The list Plain ships with contains 20000 common
    passwords (lowercased and deduplicated), created by Royce Williams:
    https://gist.github.com/roycewilliams/226886fd01572964e1431ac8afc999ce
    The password list must be lowercased to match the comparison in validate().
    """

    @cached_property
    def DEFAULT_PASSWORD_LIST_PATH(self) -> Path:
        return Path(__file__).resolve().parent / "common-passwords.txt.gz"

    def __init__(self, password_list_path: Path | None = None) -> None:
        resolved_path: Path = (
            password_list_path
            if password_list_path
            else self.DEFAULT_PASSWORD_LIST_PATH
        )
        try:
            with gzip.open(resolved_path, "rt", encoding="utf-8") as f:
                self.passwords = {x.strip() for x in f}
        except OSError:
            with open(resolved_path) as f:
                self.passwords = {x.strip() for x in f}

    def __call__(self, password: str) -> None:
        if password.lower().strip() in self.passwords:
            raise ValidationError(
                "This password is too common.",
                code="password_too_common",
            )


@deconstructible
class NumericPasswordValidator:
    """
    Validate that the password is not entirely numeric.
    """

    def __call__(self, password: str) -> None:
        if password.isdigit():
            raise ValidationError(
                "This password is entirely numeric.",
                code="password_entirely_numeric",
            )
